from datetime import datetime

import pytz


def time_str(time=None, timezone=None, Ymd_sep="-", HMS_sep=":", date_time_sep=" "):
    # 获取当前时间
    if not time:
        time = datetime.now()

    if not timezone:
        timezone = pytz.timezone('Asia/Shanghai')

    # 设置时区为+8（东八区）
    time = time.astimezone(timezone)

    # 将当前时间转换为字符串
    time = time.strftime(f"%Y{Ymd_sep}%m{Ymd_sep}%d{date_time_sep}%H{HMS_sep}%M{HMS_sep}%S")

    return time



import os
from typing import Optional, List, Literal, Union

import cv2
import imagesize
import numpy


def get_image_size(image_path):
    """
    :param image_path: 图像路径
    :return: height, width
    """

    width, height = imagesize.get(image_path)

    if (height, width) == (-1, -1):
        image: numpy.ndarray = cv2.imread(image_path, cv2.IMREAD_GRAYSCALE)
        # 获取图像尺寸
        height, width = image.shape

    return height, width


def read_image(image_path, grey: bool = False, resize=None, rgb_mode: bool = True):
    """
    读取图像函数

    :param image_path: 图像路径
    :param grey: 是否将图像转换为灰度图像，默认为False
    :param resize: (h, w)，默认为None
    :param rgb_mode: 使用opencv读出的图像是BGR格式，是否将图像转换为RGB格式，默认为True
    :return: 读取到的图像数组
    """

    if grey:
        # 读取灰度图像
        image = cv2.imread(image_path, cv2.IMREAD_GRAYSCALE)
    else:
        # 读取彩色图像
        image = cv2.imread(image_path)
        if rgb_mode:
            # 将BGR格式转换为RGB格式
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    if resize:
        image = cv2.resize(image, (resize[1], resize[0]))

    return image


def read_images(image_paths: List[str], grey: bool = False, resize=None, rgb_mode: bool = True):
    if not len(image_paths):
        raise ValueError("The length of image_paths should not be 0.")

    if not resize:
        resize = get_image_size(image_paths[0])
    image_list = []
    for image_path in image_paths:
        image = read_image(image_path, grey, resize, rgb_mode)
        image_list.append(image)

    return numpy.stack(image_list, axis=0)


def get_video_info(video_path):
    """
    获取帧速率、帧数、宽度和高度
    :param video_path: 视频文件地址
    :return: 帧数、宽度、高度和帧速率。如果视频文件无法打开，则返回None。
    """

    # 打开视频文件
    cap = cv2.VideoCapture(video_path)

    # 检查视频文件是否成功打开
    if not cap.isOpened():
        raise ValueError(f"Can't open video file {video_path}.")

    # 获取帧速率、帧数、宽度和高度
    frames_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    fps = int(cap.get(cv2.CAP_PROP_FPS))
    height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))

    # 释放VideoCapture对象
    cap.release()

    # 返回帧数、宽度、高度和帧速率
    return frames_count, fps, height, width


def read_video(video_path,
               *specific_frames: Optional[int],
               # start: int = 0, stop: int = -1, step: int = 1,
               resize=None,
               rgb_mode=True):
    """
    读取视频文件，并返回指定范围的帧。 \n
    运行此函数需要cv2模块进行视频文件的读取，numpy模块进行高效的数值计算。 \n
    示例：\n
    video = read_video("example.mp4", 0, 100, 200, 300, resize=(224, 224), rgb_mode=True) \n
    video = read_video("example.mp4", *range(0, 1000, 100), resize=(224, 224), rgb_mode=True) \n

    :param video_path: 视频文件的路径，字符串格式。
    :param specific_frames: 需要读取的特定帧的序列，每个元素为帧的索引，整数类型。如果该参数为空，则读取视频中的所有帧。
    :param resize: 重新设定帧的宽和高的尺寸，形式为(h, w)，默认为None，即不改变帧的尺寸。
    :param rgb_mode: 布尔值，确定是否将帧的颜色空间从BGR转换为RGB，为True时进行转换，默认为True。

    :return: 如果读取成功，返回一个numpy数组，形式为 (帧数，高，宽，颜色通道数)，即每一帧图像上包含的像素信息。如果读取失败，返回None。

    :raise ValueError: 当需要读取的特定帧的索引超出指定范围时，引发ValueError。
    """
    cap = cv2.VideoCapture(video_path)

    frames_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

    # check specific_frames
    if specific_frames:
        for frame in specific_frames:
            if frame < 0 or frame >= frames_count:
                raise ValueError(f"frame index {frame} out of range [0, {frames_count}).")
    else:
        specific_frames = range(frames_count)

    video = []
    for frame in specific_frames:
        cap.set(cv2.CAP_PROP_POS_FRAMES, frame)
        ret, frame = cap.read()
        if rgb_mode:
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        if resize:
            frame = cv2.resize(frame, (resize[1], resize[0]))
        video.append(frame)

    cap.release()

    # post-processing frames
    video = numpy.stack(video, axis=0)

    return video


def write_image(image: numpy.ndarray, filename: str, *, resize=None, rgb_mode=True) -> None:
    """
    将图像写入文件。
    :param image: numpy.array, 输入图像。
    :param filename: str, 输出文件名。
    :param resize: (h, w)，默认为None
    :param rgb_mode: cv2是以BGR的形式写入，如果当前是RGB需要转化为BGR，其他软件读取才会正常。默认为True。
    """
    if rgb_mode:
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    if resize:
        image = cv2.resize(image, (resize[1], resize[0]))
    cv2.imwrite(filename, image)

    print(f"{filename} written successfully!")


# 将文件格式及对应的fourcc编码方式放入常量中
FOURCC_MAP = {'.mp4': 'mp4v', '.avi': 'XVID', '.mkv': 'avc1',
              '.mov': 'mp4v', '.3gp': 'mp4v', '.webm': 'VP80', '.wmv': 'WMV1'}


def write_video(video: numpy.ndarray, filename: str, *, fps: int = 30, resize=None, rgb_mode=True) -> None:
    """
    将cv2读取的维度格式FHWC（frame_count, height, width, channel_count）视频帧写入文件。
    如果缺少channel_count维度，则会repeat补上。
    函数接受一个视频帧、一个输出文件名和一个视频帧率（默认为30帧/秒），

    使用OpenCV的VideoWriter函数创建视频写入对象，然后逐帧将视频写入文件。
    最后，释放视频写入对象。注意，为了使写入的视频具有更好的兼容性，我们选择了MP4编码格式。

    :param video: numpy.array, 输入视频帧。
    :param filename: str, 输出文件名。
    :param fps: int, 输出视频帧率。
    :param resize: (h, w)，默认为None
    :param rgb_mode: cv2是以BGR的形式写入，如果当前是RGB需要转化为RGB，默认为True。

    :raise Exception: 当输入视频维度不是FHW[C]时，引发异常。
    """

    video = numpy.array(video, dtype=numpy.uint8)

    if video.ndim not in [3, 4]:
        raise Exception("The input video should be a 3D or 4D numpy array, but got {}D array.".format(video.ndim))

    if video.ndim == 3:
        video = numpy.repeat(video[..., numpy.newaxis], 3, axis=-1)

    frame_count, height, width, channel_count = video.shape

    _, ext = os.path.splitext(filename)
    fourcc = cv2.VideoWriter_fourcc(*FOURCC_MAP[ext.lower()])  # 获取fourcc编码

    if resize:
        # 在resize后更新宽和高
        height, width = resize
        video = numpy.array([cv2.resize(frame, (resize[1], resize[0])) for frame in video])

    if rgb_mode:
        # 如果在RGB模式下，将所有帧一次性转换以提高效率
        video = [cv2.cvtColor(frame, cv2.COLOR_RGB2BGR) for frame in video]

    writer = cv2.VideoWriter(filename, fourcc, fps, (width, height))

    for frame in video:
        writer.write(frame)

    writer.release()

    print(filename, "written")


def keep_ratio_resize_image(image, target_size, position: Union[Literal["center", "random"],] = "center"):
    """
    根据target_size，如果图像的比例和target_size不一致，则先按照target_size的比例切割图像，然后再进行缩放
    :param position:
    :param image: numpy.array, (h, w[, c])，输入图像。
    :param target_size: (h, w)，目标尺寸。
    :return: numpy.array, 缩放后的图像。
    """
    ndim = image.ndim
    h, w = image.shape[:2]
    th, tw = target_size

    if h / w > th / tw:
        # 高度过长，需要先按照宽度切割
        new_h = int(w / tw * th)
        if position == "center":
            start = (h - new_h) // 2
            image = image[start: start + new_h, :, ...]
        elif position == "random":
            start = numpy.random.randint(0, h - new_h)
            image = image[start: start + new_h, :, ...]
        else:
            raise ValueError(f"Unknown position {position}.")
    else:
        # 宽度过长，需要先按照高度切割
        new_w = int(h / th * tw)
        if position == "center":
            start = (w - new_w) // 2
            image = image[:, start: start + new_w, ...]
        elif position == "random":
            start = numpy.random.randint(0, w - new_w)
            image = image[:, start: start + new_w, ...]
        else:
            raise ValueError(f"Unknown position {position}.")

    image = cv2.resize(image, (tw, th))

    if ndim == 3 and image.ndim == 2:
        return image[..., numpy.newaxis]
    else:
        return image


def keep_ratio_resize_video(video, target_size, position: Union[Literal["center", "random"],] = "center"):
    """
    根据target_size，如果视频的比例和target_size不一致，则先按照target_size的比例切割视频，然后再进行缩放
    :param position:
    :param video: numpy.array, (frames, h, w[, c])，输入视频。
    :param target_size: (h, w)，目标尺寸。
    :return: numpy.array, 缩放后的视频。
    """
    ndim = video.ndim
    frames, h, w = video.shape[:3]
    th, tw = target_size

    if h / w > th / tw:
        # 高度过长，需要先按照宽度切割
        new_h = int(w / tw * th)
        if position == "center":
            start = (h - new_h) // 2
            video = video[:, start: start + new_h, :]
        elif position == "random":
            start = numpy.random.randint(0, h - new_h)
            video = video[:, start: start + new_h, :]
        else:
            raise ValueError(f"Unknown position {position}.")
    else:
        # 宽度过长，需要先按照高度切割
        new_w = int(h / th * tw)
        if position == "center":
            start = (w - new_w) // 2
            video = video[:, :, start: start + new_w]
        elif position == "random":
            start = numpy.random.randint(0, w - new_w)
            video = video[:, :, start: start + new_w]
        else:
            raise ValueError(f"Unknown position {position}.")

    video = numpy.array([cv2.resize(frame, (tw, th)) for frame in video])

    if ndim == 3 and video.ndim == 2:
        return video[..., numpy.newaxis]
    elif ndim == 4 and video.ndim == 3:
        return video[..., numpy.newaxis]
    else:
        return video


def keep_ratio_resize(image_or_video, target_size, with_channel=True,
                      position: Union[Literal["center", "random"],] = "center"):
    """
    根据target_size，如果图像或视频的比例和target_size不一致，则先按照target_size的比例切割图像或视频，然后再进行缩放
    :param position: {"center", "random"}，默认为"center"，切割位置。
    :param image_or_video: numpy.array, (h, w[, c]) or (frames, h, w[, c])，输入图像或视频。
    :param target_size: (h, w)，目标尺寸。
    :param with_channel: 是否有通道维度，默认为True。
    :return: numpy.array, 缩放后的图像或视频。
    """
    if image_or_video.ndim == 2:
        return keep_ratio_resize_image(image_or_video, target_size, position)
    elif image_or_video.ndim == 3 and with_channel:
        return keep_ratio_resize_image(image_or_video, target_size, position)
    elif image_or_video.ndim == 3 and not with_channel:
        return keep_ratio_resize_video(image_or_video, target_size, position)
    elif image_or_video.ndim == 4:
        return keep_ratio_resize_video(image_or_video, target_size, position)
    else:
        raise ValueError(f"image_or_video should be 3D or 4D. Got {image_or_video.ndim}D.")


def pil_to_numpy(x):
    return numpy.array(x)


def to_grey(image_or_video: numpy.ndarray, with_channel=True, keep_dim=True):
    """
    Convert ndarray image or video to grey.
    :param keep_dim: whether to keep the channel dimension.
    :param image_or_video:  numpy.ndarray, (h, w) or (h, w, c) or (frames, h, w, c).
    :param with_channel:  whether there is a channel dimension in the input.
    :return: numpy.ndarray, (h, w) or (h, w, 1) or (frames, h, w, 1).
    """

    if image_or_video.ndim == 2:
        # 已经是灰度图像
        result_image = image_or_video
    elif image_or_video.ndim == 3 and with_channel:
        result_image = numpy.mean(image_or_video, axis=-1)
    elif image_or_video.ndim == 3 and not with_channel:
        # 已经是灰度图像
        result_image = image_or_video
    elif image_or_video.ndim == 4:
        result_image = numpy.mean(image_or_video, axis=-1)
    else:
        raise ValueError(
            f"image_or_video should be 2D (gray image), 3D (grey video or rgb image) or 4D (rgb video). "
            f"Got {image_or_video.ndim}D.")
    if keep_dim:
        return result_image[..., numpy.newaxis]
    else:
        return result_image


def cv2_binary(image):
    """
    cv2二值化，对MNIST效果很好
    :param image:
    :return:
    """
    # img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
    # image = cv2.GaussianBlur(image, (5, 5), 0)
    is_255 = numpy.max(image) > 1
    is_float = image.dtype == numpy.float32 or image.dtype == numpy.float64
    if not is_255:
        image = (image * 255).astype(numpy.uint8)
    if is_float:
        image = image.astype(numpy.uint8)
    # print(image.shape, image.dtype)
    # 高斯模糊
    image = cv2.GaussianBlur(image, (3, 3), 0)

    # 直方图均衡化
    equalized = cv2.equalizeHist(image)

    # 全局 Otsu 二值化
    _, global_otsu = cv2.threshold(equalized, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)

    # 自适应阈值
    adaptive_thresh = cv2.adaptiveThreshold(equalized, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 11, 2)

    # 结合 Otsu 和自适应阈值的结果
    combined = cv2.bitwise_and(global_otsu, adaptive_thresh)

    if not is_255:
        combined = combined / 255.0

    return combined



from torch.utils.data import Dataset


class SubDataset(Dataset):
    def __init__(self, original_dataset, available_index_list=None):
        self.original_dataset = original_dataset
        self.available_index_list = [i for i in available_index_list if i < len(self.original_dataset)]

    def __len__(self):
        return len(self.available_index_list)

    def __getitem__(self, index):
        return self.original_dataset[self.available_index_list[index]]


class LimitedDataset(Dataset):
    def __init__(self, original_dataset, limit):
        self.original_dataset = original_dataset
        self.limit = limit

    def __len__(self):
        return min(len(self.original_dataset), self.limit)

    def __getitem__(self, index):
        if index >= self.limit:
            raise IndexError
        return self.original_dataset[index]


class CachedDataset(Dataset):
    def __init__(self, original_dataset, maxsize=1024):
        self.original_dataset = original_dataset
        self._cached_getitem = lru_cache(maxsize=1024)(self._non_cached_getitem)

    def __len__(self):
        return len(self.original_dataset)

    def __getitem__(self, index):
        return self._cached_getitem(index)

    def _non_cached_getitem(self, index):
        return self.original_dataset[index]



import pandas as pd

# 创建一个空的 DataFrame
df = pd.DataFrame(columns=["fold", "stage", "epoch", "acc", "f1score", "precision", "recall",""])

# 添加一行数据
new_row = {"acc": 0.99, "epoch": 18, "fold": 4, "step": "train"}
df.loc[len(df)] = new_row
df.loc[len(df)] = new_row

print(df)



from typing import List

import matplotlib.pyplot as plt
import numpy


def easy_gradcam(model, target_layers, images, input_tensor=None):
    """
    This function is used to generate Class Activation Maps (CAM). CAM is a visualization technique used to understand
    the decision-making process of CNN.

    :param model: Model object, if not provided, the default will be the ResNet18 model.
    :param target_layers: Target layer, if not provided, the default will be the fourth layer of the ResNet18 model.
    :param images: Images that need to be visualized with CAM, the shape should be [b, h, w, c],
                    where b is the batch size, h is the height, w is the width, and c is the number of channels.
                    The value of the image should be in the range [0, 1].
    :param input_tensor: If not provided, images will be converted to a floating point tensor.
    :return: A list containing all original images, visual images, and grayscale CAM.
    """
    import torch
    from pytorch_grad_cam import GradCAM
    from pytorch_grad_cam.utils.image import show_cam_on_image
    from torchvision.models import resnet18

    if model is None:
        model = resnet18()

    if target_layers is None:
        target_layers = [model.layer4]

    if isinstance(images, list):
        images = torch.stack(images, dim=0)

    if input_tensor is None:
        input_tensor = torch.tensor(images, dtype=torch.float)

    if images.shape[1] == 1:
        # images = einops.rearrange(images, "b c h w -> b h w c")
        images = images.permute(0, 2, 3, 1)
        images = images.to("cpu").numpy()

    # pip install grad-cam
    # Construct the CAM object once, and then re-use it on many images:
    cam = GradCAM(model=model, target_layers=target_layers)

    # You can also use it within a with statement, to make sure it is freed,
    # In case you need to re-create it inside an outer loop:
    # with GradCAM(model=model, target_layers=target_layers, use_cuda=args.use_cuda) as cam:
    #   ...

    # We have to specify the target we want to generate
    # the Class Activation Maps for.
    # If targets is None, the highest scoring category
    # will be used for every image in the batch.
    # Here we use ClassifierOutputTarget, but you can define your own custom targets
    # That are, for example, combinations of categories, or specific outputs in a non standard model.

    # targets = [ClassifierOutputTarget(11),ClassifierOutputTarget(0)]
    targets = None

    # You can also pass aug_smooth=True and eigen_smooth=True, to apply smoothing.
    # print(input_tensor)
    grayscale_cam = cam(input_tensor=input_tensor, targets=targets)

    # In this example grayscale_cam has only one image in the batch:
    out_images = []
    for i in range(len(images)):
        # print(images.shape)
        # print(grayscale_cam.shape)
        visualization = show_cam_on_image(images[i], grayscale_cam[i], use_rgb=True)
        out_images.extend([images[i], visualization, grayscale_cam[i]])

    return out_images


def easy_tsne(data, labels, perplexity=30, n_iter=1000, random_state=0, dot_type: str = "scatter", figsize=(8, 8),
              fontsize=15, cmap="rainbow", show=True):
    """
    This function is used to visualize high-dimensional data in a low-dimensional space. The t-SNE algorithm is used to
    reduce the dimensionality of the data and visualize it in a two-dimensional space.


    :param data: High dimensional data that requires t-SNE visualization. if the data is a multi-dimensional array,
                it will be reshaped to a two-dimensional array.
    :param labels: Labels of the data, the shape should be [n].
    :param perplexity: The perplexity is related to the number of nearest neighbors that is used in other manifold
                        learning algorithms. Larger datasets usually require a larger perplexity. Consider selecting a
                        value between 5 and 50.
    :param n_iter: Maximum number of iterations for the optimization. Should be at least 250.
    :param random_state: Determines the random number generation for the optimization. Use an int to make the randomness
                        deterministic.
    :param dot_type: The type of visualization, "scatter" or "figure".
    :param figsize: The size of the matplotlib figure, the default is (10, 10).
    :param show: Whether to display the visualization, the default is True.
    :param fontsize: The font size of the label, the default is 15.
    :param cmap: The color map of the visualization, the default is "rainbow".
    :return: A list containing the original data and the visualization data.
    """
    import io
    import numpy
    from PIL import Image
    import matplotlib.pyplot as plt
    from sklearn.manifold import TSNE

    if isinstance(figsize, int):
        figsize = (figsize, figsize)

    data = data.reshape((data.shape[0], -1))
    tsne = TSNE(n_components=2, perplexity=perplexity, n_iter=n_iter, random_state=random_state)
    tsne_results = tsne.fit_transform(data)

    plt.figure(figsize=figsize)

    if dot_type == "scatter":
        plt.scatter(tsne_results[:, 0], tsne_results[:, 1], c=labels, cmap=cmap)
    else:
        plt.scatter(tsne_results[:, 0], tsne_results[:, 1], c="white")

        color_map = plt.get_cmap(cmap, len(set(labels)))  # 创建颜色映射，设置为10类

        for i in range(tsne_results.shape[0]):
            plt.text(tsne_results[i, 0], tsne_results[i, 1], str(labels[i]), c=color_map(labels[i]), fontsize=fontsize)

    plt.tight_layout()

    if show:
        plt.show()
        plt.close()
    else:
        image = io.BytesIO()

        plt.savefig(image, format='png')
        plt.close()

        image.seek(0)
        image = Image.open(image)

        return numpy.array(image)


def show_images(*images: List[numpy.array], subtitles=None, col_num: int = 3, title: str = None,
                each_size=(8, 7), show: bool = True, save_path: str = None):
    """
    Display or save a collection of images with optional subtitles.

    :param images: List of images to be displayed or saved.
    :param subtitles: List of subtitles for the images. If the number of subtitles is less than the number of images, None will be appended to the subtitles list.
    :param col_num: Number of columns for the image grid. Default is 3.
    :param title: Title for the plot. Default is None.
    :param each_size: Size for each subplot in the grid. Default is (8, 7).
    :param show: Boolean indicating whether to display the plot. Default is True.
    :param save_path: Path to save the plot. If both show and save_path are None, a ValueError is raised. Default is None.
    :return: None
    """
    if not show and save_path is None:
        raise ValueError("Must specify an output method, either show or save to a file.")

    images_count = len(images)
    col_num = min(col_num, images_count)

    each_size = (each_size, each_size * 0.875) if isinstance(each_size, int) else each_size

    subtitles = [None] * images_count if subtitles is None else subtitles + [None] * (images_count - len(subtitles))

    row_num = (images_count + col_num - 1) // col_num
    plt.subplots_adjust(top=0.85)  # 调整顶部间距
    fig, axs = plt.subplots(nrows=row_num, ncols=col_num, squeeze=False)
    if title is not None:
        fig.suptitle(title, fontsize=each_size[0] * 4)

    fig.set_size_inches(each_size[0] * col_num, each_size[1] * row_num)

    for ax, image, subtitle in zip(axs.ravel(), images, subtitles):
        ax.set_title(subtitle, fontsize=each_size[0] * 3)
        ax.imshow(image, cmap='gray' if len(image.shape) == 2 else None)
        ax.axis('off')

    for ax in axs.ravel()[images_count:]:
        ax.axis('off')

    fig.tight_layout()

    if save_path:
        fig.savefig(save_path)

    if show:
        plt.show()



import numpy
import torch
from torch import Tensor


def numpy_to_tensor(image_or_video, with_channel=True):
    """
    Convert numpy.ndarray to torch.Tensor.
    If the input is 2D with shape (h, w), the output will be (h, w).
    If the input is 3D with shape (h, w, c), the output will be (c, h, w).
    If the input is 4D with shape (frames, h, w, c), the output will be (frames, c, h, w).
    :param with_channel: whether there is a channel dimension in the input.
    :param image_or_video: numpy.ndarray. (h, w) or (h, w, c) or (frames, h, w, c).
    :return: torch.Tensor with shape (c, h, w) or (frames, c, h, w).
    """
    if isinstance(image_or_video, numpy.ndarray):
        if image_or_video.ndim == 2:
            return torch.from_numpy(image_or_video).float()
        elif image_or_video.ndim == 3 and with_channel:
            return torch.from_numpy(image_or_video.transpose((2, 0, 1))).float()
        elif image_or_video.ndim == 3 and not with_channel:
            return torch.from_numpy(image_or_video).float()
        elif image_or_video.ndim == 4:
            return torch.from_numpy(image_or_video.transpose((3, 0, 1, 2))).float()
        else:
            raise ValueError(f"image_or_video should be 3D or 4D. Got {image_or_video.ndim}D.")
    else:
        raise TypeError(f"pic should be ndarray. Got {type(image_or_video)}")


def tensor_to_numpy(image_or_video_tensor: Tensor, with_channel=True):
    """
    Convert torch.Tensor to numpy.ndarray.
    If the input is 2D with shape (h, w), the output will be (h, w).
    If the input is 3D with shape (c, h, w), the output will be (h, w, c).
    If the input is 4D with shape (frames, c, h, w), the output will be (frames, h, w, c).
    :param with_channel: whether there is a channel dimension in the input.
    :param image_or_video_tensor: torch.Tensor with shape (c, h, w) or (frames, c, h, w).
    :return: numpy.ndarray with shape (h, w, c) or (frames, h, w, c).
    """
    image_or_video_tensor = image_or_video_tensor.detach()
    if isinstance(image_or_video_tensor, Tensor):
        image_or_video_tensor = image_or_video_tensor.cpu()
        if image_or_video_tensor.dim() == 2:
            return image_or_video_tensor.numpy()
        elif image_or_video_tensor.dim() == 3 and with_channel:
            return image_or_video_tensor.permute(1, 2, 0).numpy()
        elif image_or_video_tensor.dim() == 3 and not with_channel:
            return image_or_video_tensor.numpy()
        elif image_or_video_tensor.dim() == 4:
            return image_or_video_tensor.permute(1, 2, 3, 0).numpy()
        else:
            raise ValueError(f"image_or_video_tensor should be 3D or 4D. Got {image_or_video_tensor.dim()}D.")
    else:
        raise TypeError(f"pic should be Tensor. Got {type(image_or_video_tensor)}")



import functools
import hashlib
import os
import pickle
import re
import time
from collections import defaultdict
from os import path
from typing import Literal

import numpy
import torch


def flatten(o):
    """
    This function is used to flatten a nested iterable (like a list or tuple) into a single, one-dimensional iterable.
    It uses Python's generator syntax and can handle any depth of nesting, including cases where the iterable contains strings.

    :param o: An iterable that can be nested to any depth. For example, [1, [2, [3, 4], 5]].
    :return: A generator that yields each item in the original iterable, in the order they appear, but flattened to a single level.
    """

    for item in o:
        if isinstance(item, str):
            yield item
            continue
        try:
            yield from flatten(item)
        except TypeError:
            yield item


def get_current_timestamp():
    timestamp = time.time()
    return timestamp


def equals_re(string_a, string_b, pattern_a='.*', pattern_b='.*'):
    """
    This function is used to compare two strings based on a regular expression pattern.
    The default pattern is '(.*)', which matches any part.

    :param string_a: The first string to compare.
    :param string_b: The second string to compare.
    :param pattern_a: The regular expression pattern to use for comparison. Default is '(.*)'.
    :param pattern_b: The regular expression pattern to use for comparison. Default is '(.*)'.
    :return: True if the strings match based on the pattern, False otherwise.
    """
    # print(string_a, string_b)
    # print(re.findall(pattern_a, str(string_a)), re.findall(pattern_b, str(string_b)))
    return re.findall(pattern_a, str(string_a)) == re.findall(pattern_b, str(string_b))


def join_re(list_a, list_b, key_a=0, key_b=0, how: Literal["inner", "left", "right", "outer"] = 'inner',
            pattern_a="(.*)", pattern_b="(.*)", keep_key=Literal["all", "a", "b"]):
    """
    This function is used to join two lists of tuples based on a common key.
    The function supports inner, left, right, and outer joins.
    The default join type is inner join.

    :param list_a: A list of tuples.
    :param list_b: Another list of tuples.
    :param key_a: The index of the key in list_a.
    :param key_b: The index of the key in list_b.
    :param how: The type of join to perform. Default is 'inner'.
    :param pattern_a: The regular expression pattern to use for comparison. Default is '(.*)'.
    :param pattern_b: The regular expression pattern to use for comparison. Default is '(.*)'.
    :param keep_key: The key to keep when joining. Default is 'a'.
    :return: A list of tuples that are the result of the join operation.
    """
    result = []
    if not (isinstance(list_a[0], tuple) or isinstance(list_a[0], list) or isinstance(list_a[0], numpy.ndarray)):
        list_a = [(a,) for a in list_a]

    # list_a = tuple(list_a)

    if not (isinstance(list_b[0], tuple) or isinstance(list_b[0], list) or isinstance(list_b[0], numpy.ndarray)):
        list_b = [(b,) for b in list_b]

    # list_b = tuple(list_b)

    # print(list_a)
    # print(list_b)

    if how == 'inner':
        for a in list_a:
            for b in list_b:
                if equals_re(a[key_a], b[key_b], pattern_a, pattern_b):
                    if keep_key == "a":
                        result.append([*a, *b[:key_b], *b[key_b + 1:]])
                    elif keep_key == "b":
                        result.append([*a[:key_a], *a[key_a + 1:], *b])
                    else:
                        result.append([*a, *b])
    elif how == 'left':
        for a in list_a:
            for b in list_b:
                if equals_re(a[key_a], b[key_b], pattern_a, pattern_b):
                    if keep_key == "a":
                        result.append([*a, *b[:key_b], *b[key_b + 1:]])
                    elif keep_key == "b":
                        result.append([*a[:key_a], *a[key_a + 1:], *b])
                    else:
                        result.append([*a, *b])
                    break
            else:  # no break
                result.append([*a, *(None,) * (len(list_b[0]) - 1)])
    elif how == 'right':
        for b in list_b:
            for a in list_a:
                if equals_re(a[key_a], b[key_b], pattern_a, pattern_b):
                    if keep_key == "a":
                        result.append([*a, *b[:key_b], *b[key_b + 1:]])
                    elif keep_key == "b":
                        result.append([*a[:key_a], *a[key_a + 1:], *b])
                    else:
                        result.append([*a, *b])
                    break
            else:  # no break
                result.append([*(None,) * len(list_a[0]), *b])

    elif how == 'outer':
        for a in list_a:
            for b in list_b:
                if equals_re(a[key_a], b[key_b], pattern_a, pattern_b):
                    if keep_key == "a":
                        result.append([*a, *b[:key_b], *b[key_b + 1:]])
                    elif keep_key == "b":
                        result.append([*a[:key_a], *a[key_a + 1:], *b])
                    else:
                        result.append([*a, *b])
                    break
            else:  # no break
                result.append([*a, *(None,) * (len(list_b[0]) - 1)])

        for b in list_b:
            for a in list_a:
                if equals_re(a[key_a], b[key_b], pattern_a, pattern_b):
                    break
            else:  # no break
                result.append([*(None,) * (len(list_a[0]) - 1), *b])

    return result


def group_re(file_paths, pattern="(.*)", listify=False):
    # 创建一个defaultdict来按照id进行分组
    grouped_files = defaultdict(list)

    # 按照id进行分组
    for file_path in file_paths:
        match = re.search(pattern, file_path)
        if match:
            id = match.group(1)
            grouped_files[id].append(file_path)

    if listify:
        grouped_files = list(grouped_files.items())
    else:
        grouped_files = dict(grouped_files)

    return grouped_files


def file_cache(cache_dir, hash=False):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Create a hash of the function name and arguments to use as the cache filename
            if hash:
                cache_key = hashlib.md5(pickle.dumps((func.__name__, args, kwargs))).hexdigest()
                cache_path = path.join(cache_dir, f"{cache_key}.pkl")
            else:
                args_filename = re.sub(r'[<>:"/\\|?*\x00-\x1F]', '_', f"args={args}, kw={kwargs}")
                args_filename = re.sub(r'(^\.|(\s+\.|\.\s+)$)', '', args_filename)
                cache_path = path.join(cache_dir,
                                       f"{func.__name__}({args_filename}).pkl")

            # If cache file exists, read from it; otherwise, call the original function and save the result to the cache
            if path.exists(cache_path):
                with open(cache_path, 'rb') as f:
                    result = pickle.load(f)
            else:
                result = func(*args, **kwargs)
                os.makedirs(cache_dir, exist_ok=True)
                with open(cache_path, 'wb') as f:
                    pickle.dump(result, f)

            return result

        return wrapper

    return decorator


# def extend_list_get_set(alist, get_method=lambda x: x, set_method=lambda x: x):
#     class ListWithMethod(list):
#         def __getitem__(self, item):
#             return get_method(alist[item])
#
#         def __setitem__(self, key, value):
#             alist[key] = set_method(value)
#
#     return ListWithMethod()


class extend_list_get_set(list):
    def __init__(self, alist, get_method=lambda x: x, set_method=lambda x: x):
        self.get_method = get_method
        self.set_method = set_method
        super().__init__(alist)

    def __getitem__(self, item):
        return self.get_method(super().__getitem__(item))

    def __setitem__(self, key, value):
        super().__setitem__(key, self.set_method(value))


def normalize(tensor: torch.Tensor):
    """
    Normalize tensor to [0, 1].
    :param tensor: torch.Tensor.
    :return: torch.Tensor.
    """
    tensor_min = tensor.min()
    tensor_max = tensor.max()

    return (tensor - tensor_min) / (tensor_max - tensor_min)



from setuptools import setup, find_packages


def read(filename):
    with open(filename, encoding='utf-8') as f:
        return f.read()


setup(
    name='toycv',  # 包名
    python_requires='>=3.8.0',  # python环境
    version='0.3.6',  # 包的版本
    description="A simple and flexible tool for building and training neural network models.",  # 包简介，显示在PyPI上

    long_description=read('README.md'),  # 读取的Readme文档内容，一整块字符串
    long_description_content_type="text/markdown",  # 指定包文档格式为markdown

    # 作者相关信息
    author="Yang Xiang",
    author_email='btk@qq.com',
    url='https://github.com/coderelease',

    # 指定包信息，还可以用find_packages()函数  # find_packages(where="./", include=["toycv*"]),
    # packages=["toycv",
    #           "toycv.image_video_utils"],
    packages=find_packages(where=".", include=["toycv*"]),
    install_requires=read('requirements.txt').splitlines(),  # 指定需要安装的依赖, 需要是一个列表

    # 如果你在 setup.py 中设置了 include_package_data=True，
    # 并且在 MANIFEST.in 文件中指定了 include data/*，
    # 那么在安装包时，config.ini 和 template.html 将会被自动包含在安装的包中。
    include_package_data=True,

    license="MIT",

    keywords=['toycv'],

    classifiers=[  # 一些网站的分类信息，方便用户检索
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
    ])

"""
rm -rf ./dist
pipreqs ./ --encoding=utf8 --force --mode no-pin
python3 setup.py sdist 
twine upload --repository testpypi dist/*
pip install -i https://test.pypi.org/simple/ -U toycv

twine upload --repository pypi dist/*

pip install .
"""



import os
import random

import numpy
import torch


def get_devices():
    devices = []
    backends = ["cuda", "mps", "ipu", "xpu", "mkldnn", "opengl", "opencl", "ideep", "hip", "ve", "fpga", "ort", "xla",
                "lazy", "vulkan", "meta", "hpu", "mtia", "privateuseone"]
    for backend in backends:
        if hasattr(torch.backends, backend):
            print(backend, getattr(torch, "has_" + backend))
            if getattr(torch, "has_" + backend):
                print("has_" + backend, getattr(torch, "has_" + backend))
                devices.append(backend)

    devices.append("cpu")
    return devices


def fix_seed(seed=18):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    numpy.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False



import glob
import os
import queue
import re


def ensure_dirs(file_path: str, file_has_suffix: str = True) -> str:
    if file_has_suffix and "." in os.path.basename(file_path):
        directory = os.path.dirname(file_path)
    else:
        directory = file_path

    directory = os.path.abspath(directory)

    if not os.path.exists(directory):
        os.makedirs(directory)

    return file_path


def home_path():
    return os.path.expanduser('~')


def dir_file_ext(file_path):
    return os.path.dirname(file_path), os.path.basename(file_path), os.path.splitext(file_path)[1]


def glob_re(root_dir, path_re="**"):
    """
    Match file paths based on regular expressions.

    :param root_dir: The root directory to start the search.
    :param path_re: The regular expression pattern to match the file paths. Default is "**" which matches all files.
    :return: A set of file paths that match the regular expression.
    """
    path_re_parts = path_re.split(os.path.sep)

    dir_queue = queue.SimpleQueue()
    dir_queue.put([root_dir, 0])
    result_set = set()

    while not dir_queue.empty():
        current_dir, parts_index = dir_queue.get()
        pending_paths = []

        if path_re_parts[parts_index] == "**":
            pending_paths = glob.glob(pathname=os.path.join(current_dir, "**"), recursive=True)
        else:
            pattern = re.compile(path_re_parts[parts_index])
            for listed_path in os.listdir(current_dir):
                if pattern.fullmatch(listed_path):
                    pending_paths.append(os.path.join(current_dir, listed_path))

        if parts_index == len(path_re_parts) - 1:
            result_set.update(pending_paths)
        else:
            for path in filter(os.path.isdir, pending_paths):
                dir_queue.put([path, parts_index + 1])

    # pprint(result_set)
    return result_set
